# -*- coding: utf-8 -*-
"""
Helpers for nice shell output formatting.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from . import shell


def info(msg, *args, **kw):
    """ Print sys message to stdout.

    System messages should inform about the flow of the script. This should
    be a major milestones during the build.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    shell.cprint('-- <32>{}<0>'.format(msg))


def err(msg, *args, **kw):
    """ Per step status messages

    Use this locally in a command definition to highlight more important
    information.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    shell.cprint('-- <31>{}<0>'.format(msg))



# -*- coding: utf-8 -*-
""" Commands for building and pushing the app image.

This is mainly for web apps that are deployed with docker.
"""
from __future__ import absolute_import
from . import cli, click


@cli.group('docker')
def docker_cli():
    """ Commands for building the project docker images. """
    pass


@docker_cli.command('build')
def build_images():
    """ Build and tag a docker image for the project.

    It requires a DOCKER_IMAGES conf variable to be set and contain the
    definitions for all the images built by the project. If DOCKER_REGISTRY is
    also set, then the resulting images will a have name of
    ``<registry>/<image>``. This will create two tags for each image:
    ``<name>:<version>`` and ``<name>:latest`` where <name> is the image name
    as specified in DOCKER_IMAGES entry and <version> is the current project
    version (as shown by ``peltak version``).

    Config example::

        \b
        conf.init({
            'DOCKER_REGISTRY': 'registry.mydomain.com',
            'DOCKER_IMAGES': [
                {'name': 'myapp',}
            ]
        })

    Example::

        \b
        $ peltak docker build

    """
    from peltak.core import conf
    from peltak.core import docker

    registry = conf.get('DOCKER_REGISTRY')
    docker_images = conf.get('DOCKER_IMAGES', [])

    for image in docker_images:
        docker.build_image(registry, image)


@docker_cli.command('push')
def push_images():
    """ Push project docker images to the registry.

    This command requires both DOCKER_IMAGES and DOCKER_REGISTRY conf variables
    to be set. This will push the images built by ``peltak docker build`` to the
    specified registry. For the details of how the names and tags are generated,
    see ``peltak docker build --help``.

    Config example::

        \b
        conf.init({
            'DOCKER_REGISTRY': 'registry.mydomain.com',
            'DOCKER_IMAGES': [
                {'name': 'myapp',}
            ]
        })

    Example::

        \b
        $ peltak docker push

    """
    import sys
    from peltak.core import conf
    from peltak.core import docker
    from peltak.core import log

    registry = conf.get('DOCKER_REGISTRY')
    docker_images = conf.get('DOCKER_IMAGES', [])

    if registry is None:
        log.err("You must define DOCKER_REGISTRY conf variable to push images")
        sys.exit(-1)

    for image in docker_images:
        docker.push_image(registry, image)


@docker_cli.command('list')
@click.option(
    '-p', '--password', 'registry_pass',
    type=str,
    prompt='Password',
    hide_input=True,
    help='Registry password'
)
def docker_list(registry_pass):
    """ List docker images and their tags on the remote registry.

    This command requires the DOCKER_REGISTRY conf variable to be set. You can
    also hard-code the username with the DOCKER_REGISTRY_USER conf variable. You
    will be asked for the password every time though for security reasons.

    Config example::

        \b
        conf.init({
            'DOCKER_REGISTRY': 'registry.mydomain.com',
            'DOCKER_REGISTRY_USER': 'myuser'
        })

    Example::

        \b
        $ peltak docker list            # List images in the docker registry
        $ peltak docker list -p mypass  # List registry images, use give pw

    """
    import sys
    from peltak.core import conf
    from peltak.core import docker
    from peltak.core import log
    from peltak.core import shell

    registry = conf.get('DOCKER_REGISTRY', None)

    if registry is None:
        log.err("You must define DOCKER_REGISTRY conf variable to list images")
        sys.exit(-1)

    registry_user = conf.get('DOCKER_REGISTRY_USER', None)

    if registry_user is None:
        registry_user = click.prompt("Username")

    rc = docker.RegistryClient(registry, registry_user, registry_pass)
    images = {x: rc.list_tags(x) for x in rc.list_images()}

    shell.cprint("<32>Images in <34>{} <32>registry:", registry)
    for image, tags in images.items():
        shell.cprint('  <92>{}', image)
        for tag in tags:
            shell.cprint('      <90>{}:<35>{}', image, tag)



# -*- coding: utf-8 -*-
""" Helpers for related docker commands. """
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import requests

# local imports
from . import conf
from . import log
from . import shell
from . import versioning


class RegistryClient(object):
    """ Helper class for talking to docker registry. """
    def __init__(self, registry, username, password):
        self.user = username
        self.pw = password
        self.auth = (username, password)
        self.registry_url = 'https://' + registry

    def list_images(self):
        """ List images stored in the registry.

        :return List[str]:
            List of image names.
        """
        r = requests.get(self.registry_url + '/v2/_catalog', auth=self.auth)
        return r.json()['repositories']

    def list_tags(self, image_name):
        """ List all tags for the given image stored in the registry.

        :param str image_name:
            The name of the image to query. The image must be present on the
            registry for this call to return any values.
        :return List[str]:
            List of tags for that image.
        """
        tags_url = self.registry_url + '/v2/{}/tags/list'

        r = requests.get(tags_url.format(image_name), auth=self.auth)
        data = r.json()

        if 'tags' in data:
            return reversed(sorted(data['tags']))

        return []


def build_image(registry, image):
    """ Build docker image.

    :param str registry:
        The name of the registry this image belongs to. If not given, the
        resulting image will have a name without the registry.
    :param Dict image:
        The dict containing the information about the built image. This is the
        same dictionary as defined in DOCKER_IMAGES variable.
    """
    values = {
        'registry': '' if registry is None else registry + '/',
        'image': image['name'],
        'version': versioning.current(),
    }
    args = [
        '-t {registry}{image}'.format(**values),
        '-t {registry}{image}:{version}'.format(**values),
    ]
    if 'file' in image:
        args.append('-f {}'.format(conf.proj_path(image['file'])))

    with conf.within_proj_dir(image.get('path', '.')):
        log.info("Building <33>{registry}<35>/{image}".format(**values))
        shell.run('docker build {args} . '.format(args=' '.join(args)))


def push_image(registry, image):
    """ Push the given image to selected repository.

    :param str registry:
        The name of the registry we're pushing to. This is the address of the
        repository without the protocol specification (no http(s)://)
    :param Dict image:
        The dict containing the information about the image. This is the same
        dictionary as defined in DOCKER_IMAGES variable.
    """
    values = {
        'registry': registry,
        'image': image['name'],
    }

    log.info("Pushing <33>{registry}<35>/{image}".format(**values))
    shell.run('docker push {registry}/{image}'.format(**values))



# -*- coding: utf-8 -*-
""" git helpers. """
from __future__ import absolute_import
from . import cli, click


@cli.group('git')
def git_cli():
    """ Git related commands """
    pass


@git_cli.command('add-hooks')
def add_hooks():
    """ Setup project git hooks.

    This will run all the checks before pushing to avoid waiting for the CI
    fail.

    Example::

        $ peltak git add-hooks
    """
    import os
    from peltak.core import conf
    from peltak.core import log

    log.info("Adding pre-commit hook")
    with open(conf.proj_path('.git/hooks/pre-commit'), 'w') as fp:
        fp.write('\n'.join([
            '#!/bin/bash',
            'PATH="/opt/local/libexec/gnubin:$PATH"',
            (
                'REPO_PATH=$(dirname "$(dirname "$(dirname '
                '"$(readlink -fm "$0")")")")'
            ),
            '',
            'source "$REPO_PATH/env/bin/activate"',
            '',
            'peltak lint --commit',
        ]))
        fp.write('\n')

    log.info("Adding pre-push hook")
    with open(conf.proj_path('.git/hooks/pre-push'), 'w') as fp:
        fp.write('\n'.join([
            '#!/bin/bash',
            'PATH="/opt/local/libexec/gnubin:$PATH"',
            (
                'REPO_PATH=$(dirname "$(dirname "$(dirname '
                '"$(readlink -fm "$0")")")")'
            ),
            '',
            'source "$REPO_PATH/env/bin/activate"',
            '',
            'peltak test --allow-empty',
        ]))
        fp.write('\n')

    log.info("Making hooks executable")
    os.chmod(conf.proj_path('.git/hooks/pre-commit'), 0o755)
    os.chmod(conf.proj_path('.git/hooks/pre-push'), 0o755)


@git_cli.command()
def push():
    """ Push the current branch and set to track remote.

    Example::

        $ peltak git push

    """
    from peltak.core import git
    from peltak.core import shell

    branch = git.current_branch()
    shell.run('git push -u origin {}'.format(branch))


@git_cli.command()
@click.argument('target', required=False)
def merged(target=None):
    """ Checkout the target branch, pull and delete the merged branch.

    This is to ease the repetitive cleanup of each merged branch.

    Example Config (those the are defaults)::

        \b
        conf.init({
            'MAIN_BRANCH': 'develop',
            'MASTER_BRANCH': 'master',
            'PROTECTED_BRANCHES': ['master', 'develop'],
            'RELEASE_BRANCH_PATTERN: 'release/*'
        })

    Example::

        $ peltak git merged develop # Branch was merged to develop
        $ peltak git merged master  # Branch was merged to master
        $ peltak git merged         # Autodetect where the branch was merged

    """
    from fnmatch import fnmatch
    from peltak.core import conf
    from peltak.core import git
    from peltak.core import log
    from peltak.core import shell

    main_branch = conf.get('MAIN_BRANCH', 'develop')
    master_branch = conf.get('MASTER_BRANCH', 'master')
    protected_branches = conf.get('PROTECTED_BRANCHES', ('master', 'develop'))
    release_branch_pattern = conf.get('RELEASE_BRANCH_PATTERN', 'release/*')
    branch = git.current_branch()

    if target is None:
        if fnmatch(branch, release_branch_pattern):
            target = master_branch
        else:
            target = main_branch

    try:
        shell.run('git rev-parse --verify {}'.format(branch))
    except IOError:
        log.err("Branch '{}' does not exist".format(branch))

    log.info("Checking out <33>{}".format(target))
    shell.run('git checkout {}'.format(target))

    log.info("Pulling latest changes")
    shell.run('git pull origin {}'.format(target))

    if branch not in protected_branches:
        log.info("Deleting branch <33>{}".format(branch))
        shell.run('git branch -d {}'.format(branch))

    log.info("Pruning")
    shell.run('git fetch --prune origin')

    log.info("Checking out <33>{}<32> branch".format(target))
    shell.run('git checkout {}'.format(target))



# -*- coding: utf-8 -*-
""" Time related utilities. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import time


class timed_block(object):  # noqa
    """ Context manager to measure execution time for a give block of code.

    .. code-block:: python

        with timed_block() as t:
            time.sleep(1)

        print("The block took {}ms to execute".format(t.elapsed_ms)
    """
    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed = time.time() - self.t0
        self.elapsed_s = round(self.elapsed, 3)
        self.elapsed_ms = round(self.elapsed * 1000, 3)



# -*- coding: utf-8 -*-
"""
Various commands that do not fit another category but there's not enough of them
to justify a separate module.
"""
from __future__ import absolute_import, print_function
from . import cli, click


@cli.command()
@click.option(
    '--recreate',
    is_flag=True,
    help="Delete build and out directories before running."
)
@click.option(
    '--no-index',
    is_flag=True,
    help="Do not generate main index"
)
def docs(recreate=False, no_index=False):
    """ Build project documentation.

    This command will run sphinx-refdoc first to generate the reference
    documentation for the code base. Then it will run sphinx to generate the
    final docs. You can configure the directory that stores the docs source
    (index.rst, conf.py, etc.) using the DOC_SRC_PATH conf variable. In case you
    need it, the sphinx build directory is located in ``BUILD_DIR/docs``.

    The reference documentation will be generated for all directories listed
    under 'REFDOC_PATHS conf variable. By default it is empty so no reference
    docs are generated.

    Sample Config::

        \b
        conf.init({
            'BUILD_DIR': '.build',
            'DOC_SRC_PATH': 'docs',
            'REFDOC_PATHS': [
                'src/mypkg'
            ]
        })

    Examples::

        \b
        $ peltak docs                           # Generate docs for the project
        $ peltak docs --no-index                # Skip main reference index
        $ peltak docs --recreate --no-index     # Build docs from clean slate

    """
    import os.path
    import shutil
    from peltak.core import conf
    from peltak.core import log
    from peltak.core import shell

    build_dir = conf.get_path('BUILD_DIR', '.build')
    doc_src_path = conf.get_path('DOC_SRC_PATH', 'docs')
    refdoc_paths = conf.get('REFDOC_PATHS', [])

    docs_out_path = os.path.join(doc_src_path, 'html')
    docs_ref_path = os.path.join(doc_src_path, 'ref')
    docs_assets_path = os.path.join(doc_src_path, 'assets')
    docs_build_path = os.path.join(build_dir, 'docs')

    log.info('Ensuring assets directory <94>{}<32> exists', docs_assets_path)
    if not os.path.exists(docs_assets_path):
        os.makedirs(docs_assets_path)

    if recreate:
        for path in (docs_out_path, docs_build_path):
            if os.path.exists(path):
                log.info("<91>Deleting <94>{}".format(path))
                shutil.rmtree(path)

    if refdoc_paths:
        _gen_ref_docs(docs_ref_path, no_index)
    else:
        log.err('Not generating any reference documentation - '
                'No REFDOC_PATHS specified in config')

    with conf.within_proj_dir(doc_src_path):
        log.info('Building docs with <35>sphinx')
        shell.run('sphinx-build -b html -d {build} {docs} {out}'.format(
            build=docs_build_path,
            docs=doc_src_path,
            out=docs_out_path,
        ))
        log.info('You can view the docs by browsing to <34>file://{}'.format(
            os.path.join(docs_out_path, 'index.html')
        ))


def _gen_ref_docs(ref_path, no_index=False):
    import os.path
    import shutil
    import sys
    from peltak.core import conf
    from peltak.core import log
    try:
        from refdoc import generate_docs as _generate_docs
    except ImportError:
        print("You need to install sphinx-refdoc to use the 'docs' command",
              file=sys.stderr)
        sys.exit(-1)

    refdoc_paths = conf.get('REFDOC_PATHS', [])

    log.info('Removing previously generated reference documentation')
    if os.path.exists(ref_path):
        shutil.rmtree(ref_path)

    os.makedirs(ref_path)

    log.info('Generating reference documentation')
    args = {'out_dir': ref_path}

    if no_index:
        args['no_index'] = True

    pkg_paths = [conf.proj_path(p) for p in refdoc_paths]

    _generate_docs(pkg_paths, **args)



# -*- coding: utf-8 -*-
""" Frontend related commands.

All frontend commands should be implemented through
``src/frontend/package.json``. This is just a proxy to allow easily running
them from the root project directory.
"""
from __future__ import absolute_import
from . import cli, click


def _fe_cmd(cmd):
    from peltak.core import conf
    from peltak.core import log
    from peltak.core import shell

    frontend_path = conf.get_path('FRONTEND_PATH', None)

    if frontend_path is not None:
        with conf.within_proj_dir(frontend_path):
            shell.run(cmd)
    else:
        log.err("No FRONTEND_PATH defined in the config")


@cli.command()
@click.argument('cmd')
def fe(cmd):
    """ Run a predefined frontend command.

    The commands can be defined through the FRONTEND_CMDS project configuration
    value. It should be a dict of commands mapped to the actual scripts ran
    inside the FRONTEND_DIR.

    This command is mostly helpful if frontend code base is a subdirectory of
    the project. This is a very thin wrapper around javascript task runners,
    but it allows to run those commands from within any subdirectory of the
    project (as the root is defined by ``pelconf.py`` and not ``package.json``.

    If you only run frontend commands from within the frontend directory then
    this command is probably an unnecessary overhead.

    Sample Config::

        \b
        conf.init({
            'FRONTEND_DIR': 'client/webui',
            'FRONTEND_CMDS': {
                'build': 'npm run build',
                'watch': 'npm run watch'
            }
        })

    Examples::

        \b
        $ peltak fe build       # Run frontend command named 'build'
        $ peltak fe watch       # Run frontend command named 'watch'
    """
    from peltak.core import conf
    from peltak.core import log

    frontend_cmds = conf.get('FRONTEND_CMDS', {
        'build': 'npm run build',
        'start': 'npm start',
        'watch': 'npm run watch',
        'test': 'npm test',
        'lint': 'npm lint',
        'check': 'npm run lint && npm run test'
    })

    if cmd in frontend_cmds:
        _fe_cmd(frontend_cmds[cmd])
    else:
        log.err("No {} in FRONTEND_CMDS".format(cmd))


@cli.command('init-fe')
@click.argument('cmd')
@click.option(
    '--no-recreate', 'skip_if_exists',
    is_flag=True,
    help="Do nothing if node_modules already exists"
)
def init_fe(skip_if_exists):
    """ Initialize frontend.

    This is a little shortcut for ``npm init``. The main difference is that
    by default it will not do anything if ``node_modules`` directory is already
    present. Useful for CI runs when you have your node_modules cached and you
    know nothing changed but you still want to run it if the ``node_modules``
    directory is missing.

    Config Sample::

        \b
        conf.init({
            'FRONTEND_PATH': 'client/webui'
        })

    Examples::

        \b
        $ peltak init-fe                # Run 'npm-init'
        $ peltak init-fe --no-recreate  # Run only if node_modules is missing

    """
    from os.path import exists, join
    from peltak.core import conf

    frontend_path = conf.get_path('FRONTEND_PATH', None)
    initialized = exists(join(frontend_path, 'node_modules'))

    if not initialized or not skip_if_exists:
        _fe_cmd('npm install')



# -*- coding: utf-8 -*-
""" Commands package.

All commands (and only commands) should be defined inside this package. This
should be as thin layer as possible. Ideally just processing CLI params and
displaying results.
"""
from __future__ import absolute_import

# 3rd party imports
import click

import peltak


@click.group()
@click.version_option(version=peltak.__version__, message='%(version)s')
def cli():
    """

    To get help for a specific command:

        \033[1mpeltak <command> --help\033[0m

    Examples:

        \033[1mpeltak lint --help\033[0m

        \033[1mpeltak version bump --help\033[0m

        \033[1mpeltak release upload --help\033[0m

    """
    pass



# -*- coding: utf-8 -*-
""" Commands related to Google AppEngine.

Only useful on appengine projects. If you're not using AppEngine, do not
import inside `pelconf.py`.
"""
from __future__ import absolute_import
from . import cli, click


@cli.group()
def appengine_cli():
    """ Google AppEngine related commands. """
    pass


@appengine_cli.command()
@click.option('-p', '--port', type=int, default=8080)
@click.option('--admin-port', type=int, default=None)
@click.option('--clear', is_flag=True)
def devserver(port, admin_port, clear):
    """ Run devserver. """
    from peltak.core import conf
    from peltak.core import shell

    admin_port = admin_port or (port + 1)

    args = [
        '--port={}'.format(port),
        '--admin_port={}'.format(admin_port)
    ]

    if clear:
        args += ['--clear_datastore=yes']

    with conf.within_proj_dir():
        shell.run('dev_appserver.py . {args}'.format(args=' '.join(args)))


@appengine_cli.command()
@click.option('-v', '--version', type=str, default='playground')
@click.option('--promote', is_flag=True)
def deploy(version, promote):
    """ Deploy to Google AppEngine. """
    from peltak.core import conf
    from peltak.core import shell

    with conf.within_proj_dir():
        args = [
            '-q',
            '--project=worclock',
            '--version={}'.format(version)
        ]

        if promote:
            args += ['--promote']
        else:
            args += ['--no-promote']

        shell.run('gcloud app deploy {} app.yaml '.format(' '.join(args)))


@appengine_cli.command('setup-ci')
def setup_ci():
    """ Setup AppEngine SDK on CircleCI """
    import os.path
    from peltak.core import conf
    from peltak.core import log
    from peltak.core import shell

    gcloud_path = shell.run('which gcloud', capture=True).stdout
    sdk_path = os.path.normpath(os.path.join(
        gcloud_path, '../../platform/google_appengine'
    ))
    gcloud_cmd = '/opt/google-cloud-sdk/bin/gcloud --quiet'

    if not os.path.exists(sdk_path):
        log.info("Installing AppEngine SDK")
        shell.run('sudo {} components install app-engine-python'.format(
            gcloud_cmd
        ))
    else:
        # Only initialise once. To reinitialise, just build without cache.
        log.info("AppEngine SDK already initialised")

    log.info("Using service account authentication")
    shell.run('{} auth activate-service-account --key-file {}'.format(
        gcloud_cmd,
        conf.proj_path('ops/client_secret.json')
    ))

    # shell.run('{} config set project {}'.format(gcloud_cmd, project))


def _is_appengine_sdk(path):
    """ Return True if the given *path* contains AppEngine SDK. """
    import os.path

    return all(os.path.exists(os.path.join(path, f)) for f in (
        'appcfg.py',
        'dev_appserver.py',
        'google',
        'lib',
    ))


def _get_appengine_sdk_path():
    import sys
    import os.path
    from peltak.core import shell

    sdk_path = os.environ.get('APPENGINE_SDK')

    if sdk_path is None:
        sdk_path = _find_appengine_sdk()

    if sdk_path is None:
        msg_lines = (
            '<91>AppEngine SDK not found!<0>',
            '<90>',
            ' The Google AppEngine SDK must be in your <1>$PATH<90> or you can'
            ' use  <1>$APPENGINE_SDK<90> environment variable to specify it'
            ' directly.',
            '<0>'
        )
        shell.cprint('\n'.join(msg_lines))
        sys.exit(1)

    return sdk_path


def _find_appengine_sdk():
    """ Find appengine_sdk in the current $PATH. """
    import sys
    import os.path

    paths = sys.path + os.environ.get('PATH').split(':')
    return next((path for path in paths if _is_appengine_sdk(path)), None)



# -*- coding: utf-8 -*-
""" Code linting commands. """
from __future__ import absolute_import
from . import cli, click


def _lint_files(paths, include=None, exclude=None, pretend=False):
    """ Run static analysis on the given files.

    :param paths:   Iterable with each item being path that should be linted..
    """
    from itertools import chain
    from six import string_types
    from peltak.core import conf
    from peltak.core import fs
    from peltak.core import log
    from peltak.core import shell
    from peltak.core import util

    pylint_cfg_path = conf.get_path('PYLINT_CFG_PATH', 'ops/tools/pylint.ini')
    pep8_cfg_path = conf.get_path('PEP8_CFG_PATH', 'ops/tools/pep8.ini')

    if isinstance(paths, string_types):
        raise ValueError("paths must be an array of strings")

    with util.timed_block() as t:
        files = list(chain.from_iterable(
            fs.filtered_walk(p, include, exclude) for p in paths
        ))

    log.info("Paths to lint:")
    for path in paths:
        log.info("  <0>{}", path)

    log.info("Files:")
    for p in files:
        log.info("  <0>{}", p)

    log.info("Collected <33>{} <32>files in <33>{}ms".format(
        len(files), t.elapsed_ms
    ))

    if not pretend:
        log.info("Checking PEP8 compatibility")
        files = fs.surround_paths_with_quotes(files)

        with util.timed_block() as t:
            pep8_cmd = 'pep8 --config {} {}'.format(pep8_cfg_path, files)
            pep8_ret = shell.run(pep8_cmd).return_code

            log.info("Running linter")
            pylint_cmd = 'pylint --rcfile {} {}'.format(pylint_cfg_path, files)
            pylint_ret = shell.run(pylint_cmd).return_code

        log.info("Code checked in <33>{}s", t.elapsed_s)

        if pep8_ret != 0:
            log.info("pep8 failed with return code {}", pep8_ret)

        if pylint_ret:
            log.info("pylint failed with return code {}", pylint_ret)

        return pep8_ret == pylint_ret == 0

    return True


@cli.command()
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help=("Specify patterns to exclude from linting. For multiple patterns, "
          "use the --exclude option multiple times")
)
@click.option(
    '--all', 'include_untracked',
    is_flag=True,
    help="Also include files not tracked by git."
)
@click.option(
    '--commit', 'commit_only',
    is_flag=True,
    help=("Only lint files staged for commit. Useful if you want to clean up "
          "a large code base one commit at a time.")
)
@click.option(
    '-p', '--pretend',
    is_flag=True,
    help=("Just print files that would be linted without running anything")
)
def lint(exclude, include_untracked, commit_only, pretend):
    """ Run pep8 and pylint on all project files.

    You can configure the linting paths using the LINT_PATHS config variable.
    This should be a list of paths that will be linted. If a path to a directory
    is given, all files in that directory and it's subdirectories will be
    used.

    The pep8 and pylint config paths are by default stored in ops/tools/pep8.ini
    and ops/tools/pylint.ini. You can customise those paths in your config with
    PEP8_CFG_PATH and PYLINT_CFG_PATH variables.

    **Config Example**::

        \b
        conf.init({
            'PYLINT_CFG_PATH': 'ops/tools/pylint.ini',
            'PEP8_CFG_PATH': 'ops/tools/pep8.ini',
            'LINT_PATHS': [
                'src/mypkg'
            ],
        })

    **Examples**::

        \b
        $ peltak lint               # Run linter in default mode, skip untracked
        $ peltak lint --commit      # Lint only files staged for commit
        $ peltak lint --all         # Lint all files, including untracked.
        $ peltak lint --pretend     # Print the list of files to lint
        $ peltak lint -e "*.tox*"   # Don't lint files inside .tox directory

    """
    from peltak.core import conf
    from peltak.core import git

    paths = [conf.proj_path(p) for p in conf.get('LINT_PATHS', [])]
    include = ['*.py']
    exclude = list(exclude)     # Convert from tuple to easily concatenate.

    if commit_only:
        include += ['*' + f for f in git.staged() if f.endswith('.py')]
        exclude += git.ignore()

    if not include_untracked:
        exclude += git.untracked()

    if not _lint_files(paths, include, exclude, pretend):
        exit(1)



# -*- coding: utf-8 -*-
""" Helper commands for releasing to pypi. """
from __future__ import absolute_import
from . import cli, click


@cli.group('release', invoke_without_command=True)
@click.option(
    '-c', '--component',
    type=click.Choice(['major', 'minor', 'patch']),
    required=False,
    default='patch',
    help=("Which version component to bump with this release.")
)
@click.option(
    '--exact',
    type=str,
    help="Set the newly released version to be exactly as specified."
)
@click.pass_context
def release_cli(ctx, component, exact):
    """ Create a new release branch.

    It will bump the current version number and create a release branch called
    `release/<version>` with one new commit (the version bump).

    **Example Config**::

        \b
        conf.init({
            'VERSION_FILE': 'src/mypkg/__init__.py'
        })

    **Examples**::

        \b
        $ peltak release --component=patch    # Make a new patch release
        $ peltak release -c minor             # Make a new minor release
        $ peltak release -c major             # Make a new major release
        $ peltak release                      # same as release -c patch
        $ peltak release tag                  # Tag current commit as release
        $ peltak release upload pypi          # Upload to pypi

    """
    if not ctx.invoked_subcommand:
        import os
        from peltak.core import shell
        from peltak.core import conf
        from peltak.core import log
        from peltak.core import versioning

        version_file = conf.get_path('VERSION_FILE', 'VERSION')

        with conf.within_proj_dir(quiet=True):
            out = shell.run('git status --porcelain', capture=True).stdout
            lines = out.split(os.linesep)
            has_changes = any(
                not l.startswith('??') for l in lines if l.strip()
            )

        if has_changes:
            log.info("Cannot release: there are uncommitted changes")
            exit(1)

        old_ver, new_ver = versioning.bump(component, exact)

        log.info("Bumping package version")
        log.info("  old version: <35>{}".format(old_ver))
        log.info("  new version: <35>{}".format(new_ver))

        with conf.within_proj_dir(quiet=True):
            branch = 'release/' + new_ver

            log.info("Checking out new branch <35>{}", branch)
            shell.run('git checkout -b ' + branch)

            log.info("Creating commit for the release")

            shell.run('git add {ver_file} && git commit -m "{msg}"'.format(
                ver_file=version_file,
                msg="Releasing v{}".format(new_ver)
            ))


@release_cli.command('tag')
def tag_release():
    """ Tag the current commit with as the current version release.

    This should be the same commit as the one that's uploaded as the release
    (to pypi for example).

    **Example Config**::

        \b
        conf.init({
            'VERSION_FILE': 'src/mypkg/__init__.py'
        })

    Examples::

        $ peltak release tag          # Tag the current commit as release

    """
    from peltak.core import shell
    from peltak.core import conf
    from peltak.core import git
    from peltak.core import log
    from peltak.core import versioning

    release_ver = versioning.current()
    author = git.commit_author()

    with conf.within_proj_dir(quiet=False):
        log.info("Creating tag that marks the release")
        cmd = (
            'git -c "user.name={0.name}" -c "user.email={0.email}" '
            'tag -a "v{1}" -m "Mark v{1} release"'
        ).format(
            author,
            release_ver
        )
        shell.run(cmd)


@release_cli.command()
@click.argument('target')
def upload(target):
    """ Upload to a given pypi target.

    Examples::

        \b
        $ peltak release uplaod pypi    # Upload the current release to pypi
        $ peltak release uplaod private # Upload to pypi server 'private'

    """
    from peltak.core import shell
    from peltak.core import conf
    from peltak.core import log

    log.info("Uploading to pypi server <33>{}".format(target))
    with conf.within_proj_dir(quiet=False):
        shell.run('python setup.py sdist register -r "{}"'.format(target))
        shell.run('python setup.py sdist upload -r "{}"'.format(target))


@release_cli.command('gen-pypirc')
@click.argument('username', required=False)
@click.argument('password', required=False)
def gen_pypirc(username=None, password=None):
    """
    Generate .pypirc config with the given credentials.

    Example::

        $ peltak release gen-pypirc my_pypi_user my_pypi_pass

    """
    import sys
    from os.path import join
    from peltak.core import conf
    from peltak.core import log

    path = join(conf.getenv('HOME'), '.pypirc')
    username = username or conf.getenv('PYPI_USER', None)
    password = password or conf.getenv('PYPI_PASS', None)

    if username is None or password is None:
        log.err("You must provide $PYPI_USER and $PYPI_PASS")
        sys.exit(1)

    log.info("Generating .pypirc config <94>{}".format(path))

    with open(path, 'w') as fp:
        fp.write('\n'.join((
            '[distutils]',
            'index-servers = ',
            '    pypi',
            '',
            '[pypi]',
            'repository: https://upload.pypi.org/legacy/',
            'username: {}'.format(username),
            'password: {}'.format(password),
            '',
        )))



# -*- coding: utf-8 -*-
"""
Various commands that do not fit another category but there's not enough of them
to justify a separate module.
"""
from __future__ import absolute_import
from . import cli, click


@cli.command()
@click.option(
    '-p', '--pretend',
    is_flag=True,
    help=("Just print files that would be deleted, without actually "
          "deleting them")
)
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help='Comma separated list of paths to exclude from deletion'
)
def clean(pretend, exclude):
    """ Remove temporary files like python cache, swap files, etc.

    You can configure the list of patterns with CLEAN_PATTERNS config variable.
    By default it will remove all files/dirs matching

    Config example::

        \b
        conf.init({
            'CLEAN_PATTERNS': [
                '*__pycache__*',
                '*.py[cod]',
                '*.swp'
        })

    Examples::

        \b
        $ peltak clean
        $ peltak clean -e "*.tox*"
        $ peltak clean --pretend

    """
    import os
    from os.path import isdir
    from shutil import rmtree
    from peltak.core import conf
    from peltak.core import fs
    from peltak.core import log

    clean_patterns = conf.get('CLEAN_PATTERNS', [
        '*__pycache__*',
        '*.py[cod]',
        '*.swp',
    ])

    files = fs.filtered_walk(conf.proj_path(), clean_patterns, exclude)

    for path in files:
        try:
            if not isdir(path):
                log.info('  <91>[file] <90>{}', path)
                if not pretend:
                    os.remove(path)
            else:
                log.info('  <91>[dir]  <90>{}', path)
                if not pretend:
                    rmtree(path)
        except OSError:
            log.info("<33>Failed to remove <90>{}", path)


@cli.command()
def init():
    """ Create new peltak config file in the current directory.

    If ``pelconf.py`` already exists the user will be prompted to confirm
    before continuing.

    Example::

        $ peltak init

    """
    from os.path import exists
    from peltak.core import log
    from peltak.core import shell

    config_file = 'pelconf.py'
    prompt = "-- <35>{} <32>already exists. Wipe it?<0>".format(config_file)

    if exists(config_file) and not click.confirm(shell.fmt(prompt)):
        log.info("Canceled")
        return

    log.info('Writing <35>{}'.format(config_file))
    with open('pelconf.py', 'w') as fp:
        fp.write('''# -*- coding: utf-8 -*-
""" peltak configuration file.

See https://github.com/novopl/peltak for more information.
"""
from __future__ import absolute_import

# Configure the build
from peltak.core import conf

conf.init({
})

# Import all commands
#from peltak.commands import version
''')
